from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase
from inline_snapshot._change import ExternalChange
from inline_snapshot._change import RequiredImport
from inline_snapshot._external._external_location import ExternalLocation
from inline_snapshot._external._format._protocol import get_format_handler

from ._custom import Custom


@dataclass(frozen=True)
class CustomExternal(Custom):
    value: Any
    format: str | None = None
    storage: str | None = None

    def _map(self, f):
        return f(self.value)

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        from inline_snapshot._global_state import state

        storage_name = self.storage or state().config.default_storage

        format = get_format_handler(self.value, self.format or "")

        location = ExternalLocation(
            storage=storage_name,
            stem="",
            suffix=self.format or format.suffix,
            filename=Path(context.file.filename),
            qualname=context.qualname,
        )

        tmp_file = state().new_tmp_path(location.suffix)

        storage = state().all_storages[storage_name]

        format.encode(self.value, tmp_file)
        location = storage.new_location(location, tmp_file)

        yield ExternalChange(
            "create",
            tmp_file,
            ExternalLocation.from_name("", context=context),
            location,
            format,
        )
        yield RequiredImport("create", context.file, "inline_snapshot", "external")

        return f"external({location.to_str()!r})"



import os
import sys
from pathlib import Path

import pytest
from executing import is_pytest_compatible
from rich.console import Console

from inline_snapshot._utils import is_relative_to

from . import _config
from ._exceptions import UsageError
from ._fix_assert import fix_assert
from ._flags import Flags
from ._get_snapshot_value import unwrap
from ._global_state import enter_snapshot_context
from ._global_state import leave_snapshot_context
from ._global_state import snapshot_env
from ._global_state import state
from ._snapshot_session import SnapshotSession
from .version import is_insider

pytest.register_assert_rewrite("inline_snapshot.extra")
pytest.register_assert_rewrite("inline_snapshot.testing._example")

if sys.version_info >= (3, 13):
    # fixes #186
    try:
        import readline  # noqa
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        # should fix #189 and #245
        pass


def pytest_addoption(parser, pluginmanager):
    group = parser.getgroup("inline-snapshot")

    group.addoption(
        "--inline-snapshot",
        metavar="(disable,short-report,report,review,create,update,trim,fix)*",
        dest="inline_snapshot",
        help="update specific snapshot values:\n"
        "disable: disable the snapshot logic\n"
        "short-report: show a short report\n"
        "report: show a longer report with a diff report\n"
        "review: allow to approve the changes interactive\n"
        "create: creates snapshots which are currently not defined\n"
        "update: update snapshots even if they are defined\n"
        "trim: changes the snapshot in a way which will make the snapshot more precise.\n"
        "fix: change snapshots which currently break your tests\n"
        "fix-assert: fix assertions which currently break your tests (insider only)\n",
    )

    config_path = Path("pyproject.toml")
    if config_path.exists():
        try:
            config = _config.read_config(config_path)
        except UsageError as e:
            raise pytest.UsageError(str(e))

        for name, value in config.shortcuts.items():
            value = ",".join(value)
            group.addoption(
                f"--{name}",
                action="store_const",
                const=value,
                dest="inline_snapshot",
                help=f'shortcut for "--inline-snapshot={value}"',
            )


categories = Flags.all().to_set()


def xdist_running(config):
    if "PYTEST_XDIST_WORKER" in os.environ:
        return True

    return (
        hasattr(config.option, "numprocesses")
        and config.option.numprocesses is not None
        and config.option.numprocesses != 0
    )


def find_pyproject(pytest_root, cwd):
    if is_relative_to(pytest_root, Path.cwd()):
        directory = Path.cwd()
    else:
        directory = pytest_root

    while not (pyproject_pytest := directory / "pyproject.toml").exists():
        if directory == directory.parent:
            break
        directory = directory.parent
    else:
        return pyproject_pytest


def is_xfail(request):
    if not "xfail" in request.keywords:
        return False
    xfail = request.keywords["xfail"]
    if xfail.args and xfail.args[0] == False:
        return False
    return True


@pytest.fixture(autouse=True)
def snapshot_check(request):
    SnapshotSession.test_enter()
    try:
        if is_xfail(request):
            with snapshot_env() as local_state:
                local_state.active = False
                yield
            return
        else:
            yield
    finally:
        SnapshotSession.test_exit(lambda message: pytest.fail(message, pytrace=False))


class InlineSnapshotPlugin:

    def __init__(self):
        self.session = SnapshotSession()

    @pytest.hookimpl(tryfirst=True)
    def pytest_plugin_registered(self, plugin, manager):
        """Register @customize hooks from conftest.py files"""
        # Skip internal plugins

        if not hasattr(plugin, "__file__") or plugin.__file__ is None:
            return

        # Only process conftest.py files
        if "conftest.py" not in plugin.__file__:
            return

        self.session.register_customize_hooks_from_module(plugin)

    @pytest.hookimpl
    def pytest_configure(self, config):
        enter_snapshot_context()

        # Register customize hooks from all already loaded conftest.py files
        for plugin in config.pluginmanager.get_plugins():
            if (
                hasattr(plugin, "__file__")
                and plugin.__file__
                and "conftest.py" in plugin.__file__
            ):
                self.session.register_customize_hooks_from_module(plugin)

        # setup default flags
        if is_pytest_compatible():
            state().config.default_flags_tui = ["create", "review"]
            state().config.default_flags = ["report"]

        pyproject = find_pyproject(config.rootpath, Path.cwd())

        if config.option.inline_snapshot is not None:
            cli_flags = config.option.inline_snapshot.split(",")
            cli_flags = {flag for flag in cli_flags if flag}
        else:
            cli_flags = None

        def error(message):
            raise pytest.UsageError(message)

        self.session.load_config(
            pyproject,
            cli_flags=cli_flags,
            parallel_run=xdist_running(config),
            error=error,
            project_root=config.rootpath,
        )

    @pytest.hookimpl(trylast=True)
    def pytest_sessionfinish(self, session, exitstatus):
        config = session.config

        capture = config.pluginmanager.getplugin("capturemanager")

        suspend_capture = (
            capture._global_capturing is not None
            and capture._global_capturing.in_ is not None
            and capture._global_capturing.in_._state == "started"
        )

        if suspend_capture:
            capture._global_capturing.in_.suspend()

        try:
            self.session.show_report(
                Console(
                    highlight=False,
                    no_color=config.getoption("color") == "no",
                    color_system=(
                        "standard" if config.getoption("color") == "yes" else "auto"
                    ),
                )
            )

        finally:
            if suspend_capture:
                capture._global_capturing.in_.resume()

    @pytest.hookimpl
    def pytest_unconfigure(self, config):
        leave_snapshot_context()

    @pytest.hookimpl
    def pytest_assertrepr_compare(self, config, op, left, right):

        if is_insider and state().active:
            import inspect

            frame = inspect.currentframe()
            frame = frame.f_back

            while inspect.getmodule(frame).__name__.startswith(("pluggy.", "_pytest.")):
                frame = frame.f_back

            fix_assert(frame, left, right)

        results = []
        left, left_unwrapped = unwrap(left)
        right, right_unwrapped = unwrap(right)

        if left_unwrapped or right_unwrapped:
            results = config.hook.pytest_assertrepr_compare(
                config=config, op=op, left=left, right=right
            )

        if results:
            return results[0]


@pytest.hookimpl
def pytest_configure(config):
    config.pluginmanager.register(InlineSnapshotPlugin(), "inline-snapshot-plugin")



from __future__ import annotations

from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase
from inline_snapshot._sentinels import undefined

from ._custom import Custom


class CustomUndefined(Custom):
    def __init__(self):
        self.value = undefined

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        yield from ()
        return "..."

    def _map(self, f):
        return f(undefined)



from typing import Callable

from rich.console import Console


def raise_problem(message):
    from inline_snapshot._global_state import state

    state().all_problems.add(message)


def report_problems(console: Callable[[], Console]):

    from inline_snapshot._global_state import state

    if not state().all_problems:
        return
    console().rule("[red]Problems")
    for problem in sorted(state().all_problems):
        console().print(f"{problem}")
        console().print()



from inline_snapshot._external._diff import BinaryDiff
from inline_snapshot._external._diff import TextDiff

from ._code_repr import HasRepr
from ._code_repr import customize_repr
from ._exceptions import UsageError
from ._external._external import external
from ._external._external_file import external_file
from ._external._format._protocol import Format
from ._external._format._protocol import register_format
from ._external._format._protocol import register_format_alias
from ._external._outsource import outsource
from ._get_snapshot_value import get_snapshot_value
from ._inline_snapshot import snapshot
from ._is import Is
from ._types import Category
from ._types import Snapshot
from ._unmanaged import declare_unmanaged
from .version import __version__

__all__ = [
    "snapshot",
    "external",
    "outsource",
    "customize_repr",
    "HasRepr",
    "Is",
    "Category",
    "Snapshot",
    "UsageError",
    "register_format_alias",
    "register_format",
    "Format",
    "TextDiff",
    "BinaryDiff",
    "external_file",
    "declare_unmanaged",
    "get_snapshot_value",
    "__version__",
]



from __future__ import annotations

import ast
import importlib
from dataclasses import dataclass
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase
from inline_snapshot._change import RequiredImport
from inline_snapshot._code_repr import HasRepr
from inline_snapshot._code_repr import value_code_repr
from inline_snapshot._utils import clone

from ._custom import Custom


@dataclass(frozen=True)
class Import:
    """Represents an import statement: `import module`"""

    module: str


@dataclass(frozen=True)
class ImportFrom:
    """Represents a from-import statement: `from module import name`"""

    module: str
    name: str


def _simplify_module_path(module: str, name: str) -> str:
    """Simplify module path by finding the shortest import path for a given name."""
    value = getattr(importlib.import_module(module), name)
    parts = module.split(".")
    while len(parts) >= 2:
        if getattr(importlib.import_module(".".join(parts[:-1])), name, None) == value:
            parts.pop()
        else:
            break
    return ".".join(parts)


class CustomCode(Custom):
    _imports: list[Import | ImportFrom]

    def __init__(self, value, repr_str=None, imports: list[Import | ImportFrom] = []):
        assert not isinstance(value, Custom)
        value = clone(value)
        self._imports = list(imports)

        if repr_str is None:
            self.repr_str = value_code_repr(value)

            try:
                ast.parse(self.repr_str)
            except SyntaxError:
                self.repr_str = HasRepr(type(value), self.repr_str).__repr__()
                self._imports.append(ImportFrom("inline_snapshot", "HasRepr"))
        else:
            self.repr_str = repr_str

        self.value = value

        super().__init__()

    def _map(self, f):
        return f(self.value)

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        for imp in self._imports:
            if isinstance(imp, Import):
                yield RequiredImport(flag="fix", file=context.file, module=imp.module)
            elif isinstance(imp, ImportFrom):
                yield RequiredImport(
                    flag="fix",
                    file=context.file,
                    module=_simplify_module_path(imp.module, imp.name),
                    name=imp.name,
                )
            else:
                assert False

        return self.repr_str

    def __repr__(self):
        return f"CustomCode({self.repr_str!r})"



from __future__ import annotations

import hashlib
import shutil
import typing
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from .._external_location import ExternalLocation
from ._protocol import StorageLookupError
from ._protocol import StorageProtocol


def file_digest(file: typing.BinaryIO, name: str):
    algo = hashlib.new(name)
    algo.update(file.read())
    return algo


class HashStorage(StorageProtocol):
    def __init__(self, directory):
        self.directory = Path(directory)

    def _ensure_directory(self):
        self.directory.mkdir(exist_ok=True, parents=True)
        gitignore = self.directory / ".gitignore"
        if not gitignore.exists():
            gitignore.write_bytes(
                b"# ignore all snapshots which are not referred in the source\n*-new.*\n"
            )

    @contextmanager
    def load(self, location: ExternalLocation) -> Generator[Path]:
        path = self._lookup_path(location.path)
        yield path

    def store(self, location: ExternalLocation, file_path: Path):
        self._ensure_directory()

        with file_path.open("rb") as f:
            hash_name = file_digest(f, "sha256").hexdigest()

        assert location.suffix

        if not (self.directory / (hash_name + location.suffix)).exists():
            shutil.copy(
                str(file_path),
                str(self.directory / (hash_name + location.suffix)),
            )

    def delete(self, location: ExternalLocation):
        path = self._lookup_path(location.path)
        path.unlink()

    def new_location(
        self, location: ExternalLocation, file_path: Path
    ) -> ExternalLocation:
        from inline_snapshot._global_state import state

        with file_path.open("rb") as f:
            hash_name = file_digest(f, "sha256").hexdigest()

        path = hash_name[: state().config.hash_length]
        if len(path) < len(hash_name):
            path += "*"

        return location.with_stem(path)

    def find_unused_externals(
        self, used_externals: list[ExternalLocation]
    ) -> list[ExternalLocation]:
        unused_externals = self.list()
        for location in used_externals:
            if location.path:
                used = self.lookup_all(location.path)
                unused_externals -= used

        return [ExternalLocation.from_name(name) for name in unused_externals]

    def list(self) -> set[str]:

        if self.directory.exists():
            return {item.name for item in self.directory.iterdir()} - {".gitignore"}
        else:
            return set()

    def _lookup_path(self, name) -> Path:
        if "*" not in name:
            p = Path(name)
            name = p.stem + "*" + p.suffix
        files = list(self.directory.glob(name))

        if len(files) > 1:
            raise StorageLookupError(
                f"hash collision files={sorted(f.name for f in  files)}", files=files
            )

        if not files:
            raise StorageLookupError(
                f"hash {name!r} is not found in the HashStorage", files=[]
            )

        return files[0]

    def lookup_all(self, name: str) -> set[str]:
        return {file.name for file in self.directory.glob(name)}



from __future__ import annotations

from pathlib import Path

from inline_snapshot._external._diff import BinaryDiff

from ._protocol import Format
from ._protocol import register_format


@register_format
class BinaryFormat(BinaryDiff, Format[bytes]):
    "Stores bytes in `.bin` files and shows them as a hexdump."

    suffix = ".bin"

    def is_format_for(self, value: object):
        return isinstance(value, bytes)

    def encode(self, value: bytes, path: Path):
        path.write_bytes(value)

    def decode(self, path: Path) -> bytes:
        return path.read_bytes()



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from .._external_location import ExternalLocation


class StorageLookupError(Exception):
    def __init__(self, msg, files):
        super().__init__(msg)
        self.files = files


class StorageProtocol:

    @contextmanager
    def load(self, location: ExternalLocation) -> Generator[Path]:
        """
        returns the path to a file in the storage.
        """
        raise NotImplementedError

    def store(self, location: ExternalLocation, file_path: Path):
        """
        stores the file in the storage.

        Parameters:

            location: the location in the storage
            file_path: path to a temporaray file which should be stored in the Storage
        """
        raise NotImplementedError

    def delete(self, location: ExternalLocation):
        raise NotImplementedError

    def new_location(
        self, location: ExternalLocation, file_path: Path
    ) -> ExternalLocation:
        """
        creates or changes the location where the file should be stored.
        """
        raise NotImplementedError

    def find_unused_externals(
        self, used_externals: list[ExternalLocation]
    ) -> list[ExternalLocation]:
        raise NotImplementedError

    def check_externals(self, used_externals: list[ExternalLocation]):
        """
        This function is executed with all external locations
        at the end of each test session and can be used
        to verify storage specific constraints
        """



import ast
import sys
from dataclasses import dataclass
from functools import cached_property
from types import FrameType
from typing import cast

from executing import Source

from inline_snapshot._source_file import SourceFile


@dataclass
class FrameContext:
    globals: dict
    locals: dict


class AdapterContext:

    _frame: FrameType

    def __init__(self, frame: FrameType):
        self._frame = frame

    @cached_property
    def expr(self):
        return Source.executing(self._frame)

    @property
    def source(self) -> Source:
        return cast(
            Source,
            getattr(self.expr, "source", None) if self.expr is not None else None,
        )

    @property
    def file(self):
        return SourceFile(self.source)

    @property
    def frame(self) -> FrameContext:
        return FrameContext(globals=self._frame.f_globals, locals=self._frame.f_locals)

    @cached_property
    def local_vars(self):
        """Get local vars from snapshot context."""
        return {
            var_name: var_value
            for var_name, var_value in self._frame.f_locals.items()
            if "@" not in var_name
        }

    @cached_property
    def global_vars(self):
        """Get global vars from snapshot context."""
        return {
            var_name: var_value
            for var_name, var_value in self._frame.f_globals.items()
            if "@" not in var_name
        }

    @cached_property
    def qualname(self) -> str:
        if sys.version_info >= (3, 11):
            return self._frame.f_code.co_qualname
        else:
            return self.expr.code_qualname()

    def eval(self, node):
        assert self.frame is not None

        return eval(
            compile(ast.Expression(node), self.file.filename, "eval"),
            self.frame.globals,
            self.frame.locals,
        )



from collections import namedtuple

IterResult = namedtuple("IterResult", "value list")


def split_gen(gen):
    it = iter(gen)
    l = []
    while True:
        try:
            l.append(next(it))
        except StopIteration as stop:
            return IterResult(stop.value, l)


def only_value(gen):
    return split_gen(gen).value


def gen_map(stream, f):
    stream = iter(stream)
    while True:
        try:
            yield f(next(stream))
        except StopIteration as stop:
            return stop.value


def with_flag(stream, flag):
    def map(change):
        change.flag = flag
        return change

    return gen_map(stream, map)


def make_gen_map(f):
    def m(stream):

        return gen_map(stream, f)

    return m



import ast
import os
from dataclasses import replace
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

from executing import Source

from inline_snapshot._external._external_location import ExternalLocation

from .._rewrite_code import ChangeRecorder
from .._rewrite_code import end_of
from .._rewrite_code import start_of


def contains_import(tree, module, name):
    for node in tree.body:
        if (
            isinstance(node, ast.ImportFrom)
            and node.module == module
            and any(alias.name == name for alias in node.names)
        ):
            return True
    return False


def contains_module_import(tree, module):
    for node in tree.body:
        if isinstance(node, ast.Import):
            if any(
                alias.name == module and alias.asname is None for alias in node.names
            ):
                return True
    return False


def used_externals_in(
    filename: Path, source: Union[str, ast.Module], check_import=True
) -> List[ExternalLocation]:
    if isinstance(source, str):
        if "external" not in source:
            return []
        tree = ast.parse(source)
    else:
        tree = source

    if check_import and not contains_import(tree, "inline_snapshot", "external"):
        return []

    usages = []

    for node in ast.walk(tree):
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "external"
            and node.args
            and isinstance(node.args[0], ast.Constant)
            and isinstance(node.args[0].value, str)
        ):
            arg = node.args[0].value

            try:
                location = replace(
                    ExternalLocation.from_name(arg),
                    filename=filename,
                    linenumber=node.lineno,
                )
            except ValueError:
                pass
            else:
                usages.append(location)

    return usages


def module_name_of(filename: Union[str, os.PathLike]) -> Optional[str]:
    path = Path(filename).resolve()

    assert path.suffix == ".py"

    parts = []

    if path.name != "__init__.py":
        parts.append(path.stem)

    current = path.parent

    while current != current.root:
        if not (current / "__init__.py").exists():
            break

        parts.append(current.name)

        next_parent = current.parent
        if next_parent == current:
            break  # pragma: no cover
        current = next_parent
    else:
        pass  # pragma: no cover

    parts.reverse()

    assert parts

    return ".".join(parts)


def ensure_import(
    filename,
    imports: Dict[str, Set[str]],
    module_imports: Set[str],
    recorder: ChangeRecorder,
):
    source = Source.for_filename(filename)

    change = recorder.new_change()

    tree = source.tree
    token = source.asttokens()

    my_module_name = module_name_of(filename)

    code = ""
    for module, names in imports.items():
        if module == my_module_name:
            continue
        if module == "builtins":
            continue
        for name in sorted(names):
            if not contains_import(tree, module, name):
                code += f"from {module} import {name}\n"

    for module in sorted(module_imports):
        if not contains_module_import(tree, module):
            code += f"import {module}\n"

    assert isinstance(tree, ast.Module)

    # find source position
    last_import = None
    for node in tree.body:
        if not (
            isinstance(node, (ast.ImportFrom, ast.Import))
            or (
                isinstance(node, ast.Expr)
                and isinstance(node.value, ast.Constant)
                and isinstance(node.value.value, str)
            )
        ):
            break
        last_import = node

    if last_import is None:
        position = start_of(tree.body[0].first_token)  # type: ignore
    else:
        last_token = last_import.last_token  # type: ignore
        while True:
            next_token = token.next_token(last_token)
            if last_token.end[0] == next_token.end[0]:
                last_token = next_token
            else:
                break
        position = end_of(last_token)

    if code:
        code = "\n" + code
        change.insert(position, code, filename=filename)



from __future__ import annotations

import ast
from dataclasses import dataclass
from dataclasses import field
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase

from ._custom import Custom


@dataclass(frozen=True)
class CustomDefault(Custom):
    value: Custom = field(compare=False)

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        yield from ()  # pragma: no cover
        # this should never be called because default values are never converted into code
        assert False

    def _map(self, f):
        return self.value._map(f)


def unwrap_default(value):
    if isinstance(value, CustomDefault):
        return value.value
    return value


@dataclass(frozen=True)
class CustomCall(Custom):
    node_type = ast.Call
    function: Custom = field(compare=False)
    args: list[Custom] = field(compare=False)
    kwargs: dict[str, Custom] = field(compare=False)

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        args = []
        for a in self.args:
            code = yield from a._code_repr(context)
            args.append(code)

        for k, v in self.kwargs.items():
            if not isinstance(v, CustomDefault):
                code = yield from v._code_repr(context)
                args.append(f"{k}={code}")

        return f"{yield from self.function._code_repr(context)}({', '.join(args)})"

    def argument(self, pos_or_str):
        if isinstance(pos_or_str, int):
            return unwrap_default(self.args[pos_or_str])
        else:
            return unwrap_default(self.kwargs[pos_or_str])

    def _map(self, f):
        args = [f(x._map(f)) for x in self.args]
        kwargs = {k: f(v._map(f)) for k, v in self.kwargs.items()}

        try:
            return self.function._map(f)(
                *args,
                **kwargs,
            )
        except Exception as e:
            call_args = args + [f"{k}={v}" for k, v in kwargs.items()]
            raise TypeError(
                f"can not call {self.function}({', '.join(map(str,call_args))})"
            ) from e



import typing
from typing import TYPE_CHECKING

from inline_snapshot._unmanaged import declare_unmanaged

from ._global_state import state

if TYPE_CHECKING:

    T = typing.TypeVar("T")

    def Is(v: T) -> T:
        return v

else:

    class IsMetaType(type):
        def __call__(self, value):
            if not state().active:
                return value
            return super().__call__(value)

    @declare_unmanaged
    class Is(metaclass=IsMetaType):
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return self.value == other

        def __repr__(self):
            return repr(self.value)



import io
import token
import tokenize
from pathlib import Path

from executing import Source

from inline_snapshot._format import enforce_formatting
from inline_snapshot._format import format_code
from inline_snapshot._utils import normalize
from inline_snapshot._utils import simple_token

ignore_tokens = (token.NEWLINE, token.ENDMARKER, token.NL)


def _token_of_code(code_repr):
    input = io.StringIO(code_repr)

    return [
        simple_token(t.type, t.string)
        for t in tokenize.generate_tokens(input.readline)
        if t.type not in ignore_tokens
    ]


class SourceFile:
    _source: Source

    def __init__(self, source: Source):
        self._source = source

    @property
    def filename(self) -> str:
        return self._source.filename

    def _format(self, code):
        if self._source is None or enforce_formatting():
            return code
        else:
            return format_code(code, Path(self._source.filename))

    def format_expression(self, code: str) -> str:
        return self._format(code).strip()

    def asttokens(self):
        return self._source.asttokens()

    def code_changed(self, old_node, new_code):

        if old_node is None:
            return False

        return self._token_of_node(old_node) != _token_of_code(new_code)

    def _token_of_node(self, node):

        return list(
            normalize(
                [
                    simple_token(t.type, t.string)
                    for t in self._source.asttokens().get_tokens(node)
                    if t.type not in ignore_tokens
                ]
            )
        )



from __future__ import annotations

import ast
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase

if TYPE_CHECKING:
    pass


class Custom(ABC):
    """
    Custom objects are returned by the `create_*` functions of the builder.
    They should only be returned in your customize function or used as arguments for other `create_*` functions.
    """

    node_type: type[ast.AST] = ast.AST
    original_value: Any

    def __hash__(self):
        return hash(self._eval())

    def __eq__(self, other):
        assert isinstance(other, Custom)
        return self._eval() == other._eval()

    @abstractmethod
    def _map(self, f):
        raise NotImplementedError()

    @abstractmethod
    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        raise NotImplementedError()

    def _eval(self):
        return self._map(lambda a: a)



from __future__ import annotations

import inspect
import warnings
from contextlib import contextmanager
from functools import singledispatch
from typing import TYPE_CHECKING
from unittest import mock

from typing_extensions import deprecated

from inline_snapshot._generator_utils import only_value

if TYPE_CHECKING:
    from inline_snapshot._adapter_context import AdapterContext


real_repr = repr


def adapter_context_for_parent_frame():
    from inline_snapshot._adapter_context import AdapterContext

    frame = inspect.currentframe()
    assert frame
    frame = frame.f_back
    assert frame
    frame = frame.f_back
    assert frame

    return AdapterContext(frame)


class HasRepr:
    """This class is used for objects where `__repr__()` returns an non-
    parsable representation.

    HasRepr uses the type and repr of the value for equal comparison.

    You can change `__repr__()` to return valid python code or use
    `@customize_repr` to customize repr which is used by inline-
    snapshot.
    """

    def __init__(self, type, str_repr: str) -> None:
        self._type = type
        self._str_repr = str_repr

    def __repr__(self):
        return f"HasRepr({self._type.__qualname__}, {self._str_repr!r})"

    def __eq__(self, other):

        if isinstance(other, HasRepr):
            if other._type is not self._type:
                return False
        else:
            if type(other) is not self._type:
                return False

        with mock_repr(adapter_context_for_parent_frame()):
            other_repr = value_code_repr(other)
        return other_repr == self._str_repr or other_repr == real_repr(self)


@singledispatch
def code_repr_dispatch(value):
    return real_repr(value)


@deprecated("use @customize instead")
def customize_repr(f):
    """Register a function which should be used to get the code representation
    of a object.

    ``` python
    @customize_repr
    def _(obj: MyCustomClass):
        return f"MyCustomClass(attr={repr(obj.attr)})"
    ```

    it is important to use `repr()` inside the implementation, because it is mocked to return the code representation

    you dont have to provide a custom implementation if:
    * __repr__() of your class returns a valid code representation,
    * and __repr__() uses `repr()` to get the representation of the child objects
    """
    warnings.warn(
        "@customize_repr is deprecated, @customize should be used instead",
        DeprecationWarning,
    )
    code_repr_dispatch.register(f)


def code_repr(obj):
    with mock_repr(adapter_context_for_parent_frame()):
        return repr(obj)


@contextmanager
def mock_repr(context: AdapterContext):
    assert context is not None

    def new_repr(obj):
        from inline_snapshot._customize._builder import Builder

        return only_value(
            Builder(_snapshot_context=context)._get_handler(obj)._code_repr(context)
        )

    with mock.patch("builtins.repr", new_repr):
        yield


def value_code_repr(obj):
    assert repr is not real_repr, "@mock_repr is missing"

    if not type(obj) == type(obj):  # pragma: no cover
        # this was caused by https://github.com/samuelcolvin/dirty-equals/issues/104
        # dispatch will not work in cases like this
        return (
            f"HasRepr({repr(type(obj))}, '< type(obj) can not be compared with == >')"
        )

    result = code_repr_dispatch(obj)

    return result



try:
    import dirty_equals  # type: ignore
except ImportError:  # pragma: no cover

    def is_dirty_equal(value):
        return False

else:

    def is_dirty_equal(value):
        t = value if isinstance(value, type) else type(value)
        return any(x is dirty_equals.DirtyEquals for x in t.__mro__)


def update_allowed(value):
    global unmanaged_types
    return not (is_dirty_equal(value) or isinstance(value, tuple(unmanaged_types)))  # type: ignore


unmanaged_types = []


def is_unmanaged(value):
    return not update_allowed(value)


def declare_unmanaged(data_type):
    global unmanaged_types
    unmanaged_types.append(data_type)
    return data_type
